#!/bin/bash
#
# LDoms
#
# Copyright (c) 2013, 2017, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# This script is typically called by udev in response to a change
# in the device tree (addition/removal of vswitch and vnets), to set
# up the Linux Bridge/OVS configuration for the operation of vswitches and
# vnets.
# The iproute2 utility suite commands (ip/bridge) are employed in lieu of
# the deprecated ifconfig/brctl where possible. The exception being the case
# when adding/removing ports to/from a bridge, since the "master" keyword
# which is required by the ip command for these operations is not supported
# on iproute v2.6.32 delivered with OL6. See comment in bridge_add_port()
# for an example.
# Every command that modifies the bridge configuration is logged along with
# its return code. The messages generated are logged using rsyslog(8), to
# a log file located at:
#
# /var/log/vsw.log
#
# and the entries are tagged with "VSW-LOG". The rsyslog(8) configuration,
# including the default log file location can be found at:
#
# /etc/rsyslog.d/vswlog.conf
#
# The log file size is managed with logrotate(8), using a configuration file
# located at:
#
# /etc/logrotate.d/vswlog
#

# default value for sysfs mount point - will be overridden later based on /etc/fstab
SYSFS_MNT=/sys

MDPROP=/usr/lib/ldoms/mdprop.py
LOGGER=/usr/bin/logger

LOG=/var/log/vsw.log.$$

export PATH=/sbin:/bin:/usr/sbin:/usr/bin

MAX_ATTEMPTS=5

#
# Log the execution and exit codes of commands passed as arguments.
# Returns the exit code of the command executed.
#
log_exec()
{
	local rval
	vsw_log "Running: ${1+"$@"}"
	eval ${1+"$@"}
	rval=$?
	vsw_log "Exit code: ${rval} (${1+"$@"})"
	return ${rval}
}

#
# Log errors and messages to file
#
vsw_log()
{
	local msg="$@"

	$LOGGER -p daemon.notice -t VSW-LOG -s "[${FUNCNAME[1]}] ${msg}"

	return 0
}

#
# Logs the error message and exits the script.
#
fatal_error()
{
	local msg="$1"
	vsw_log "Fatal Error: ${msg}. Exiting..."
	exit 1
}

#
# Retrieves the IP configuration of an interface. Outputs (in order):
# 1) dhcp-config
# 2) IP address in CIDR format
# 3) broadcast address
# 4) gateway routes for the interface
#
ip_get()
{
	local interface=$1
	local dhcp address broadcast

	dhcp=$(pgrep -f "dhclient.*${interface}")

	set -- $(ip addr show dev ${interface} | \
		awk '/inet .* brd/ {print $2, $4}')

	address=${1:-"0.0.0.0/0"}
	broadcast=${2:-"0.0.0.0"}

	if [[ -n "${dhcp}" ]]; then
		echo "dhcp"
	elif [[ "${address}" = "0.0.0.0/0" ]]; then
		echo "none"
	else
		echo "static"
	fi

	echo ${address}
	echo ${broadcast}

	ip route show dev ${interface} | awk '$2=="via" {print $1, $3}'
}

ip_reset()
{
	local interface=$1
	local state=$2
	local dhcp

	vsw_log "Resetting IP on ${interface}"

	dhcp=$(pgrep -f "dhclient.*${interface}")
	if [[ -n "${dhcp}" ]]; then
		kill ${dhcp}
	fi

	# Flush all IPv4 addresses from interface
	if log_exec ip -4 addr flush dev ${interface} && \
		log_exec ip link set dev ${interface} ${state}
	then
		# IP configuration has been reset for the interface
		return 0
	else
		fatal_error "Failed to reset IP configuration on ${interface}"
	fi
}

#
# Applies an IP configuration to an interface.
#
ip_set()
{
	local interface=$1
	local config=$2
	local address=$3
	local broadcast=$4
	local dest
	local gateway

	shift 4

	ip_reset ${interface} up

	if [[ "${config}" != "dhcp" && "${config}" != "static" ]]; then
		vsw_log "Configuring ${interface} with no IP"
		return 0
	fi

	if [[ -n "${address}" ]]; then
		vsw_log "Configuring ${interface} with ${address}"

		#
		# The IP address is in CIDR notation so there is no
		# need to specify the netmask
		#
		if ! log_exec ip addr add ${address} brd ${broadcast} dev ${interface}
		then
			fatal_error "Unable to assign ${address} to ${interface}"
		fi
	fi

	# add gateway routes
	while true; do
		if [[ -z "$1" || -z "$2" ]]; then
			break
		fi
		dest=${1}
		gateway=${2}
		shift 2
		vsw_log "Adding ${dest} route via ${gateway}"

		if ! log_exec ip route add ${dest} via ${gateway} dev ${interface}
		then
			fatal_error "Unable to set up route to ${dest} via ${gateway}"
		fi
	done

	if [[ "$config" = "dhcp" ]]; then
		vsw_log "Configuring ${interface} with DHCP"
		dhclient -nw ${interface} 200>&-
	else
		vsw_log "Configuring ${interface} with static configuration"
		if ! log_exec ip link set dev ${interface} up
		then
			fatal_error "Unable to set up static interface ${interface}"
		fi
	fi
}

bridge_exists()
{
	local bridge=$1

	[ -d "${SYSFS_MNT}/class/net/${bridge}" ]

	#ovs-vsctl br-exists $bridge
	#[[ $? -eq 2 ]]
}

bridge_create()
{
	local bridge=$1

	vsw_log "Creating bridge ${bridge}"

	if log_exec ip link add name ${bridge} type bridge && \
		log_exec ip link set dev ${bridge} up
	then
		return 0
	else
		fatal_error "Failed to create bridge ${bridge}"
	fi

	#ovs-vsctl add-br $bridge
}

bridge_destroy()
{
	local bridge=$1

	vsw_log "Destroying bridge ${bridge}"
	ip_reset ${bridge} down

	if ! log_exec ip link del ${bridge} type bridge
	then
		fatal_error "Unable to remove ${bridge}"
	fi

	#ovs-vsctl del-br $bridge
}

bridge_list_ports()
{
	local bridge=$1
	ls ${SYSFS_MNT}/class/net/${bridge}/brif/

	#ovs-vsctl list-ports $bridge
}

bridge_port_exists()
{
	local bridge=$1
	local port=$2

	[ -d "${SYSFS_MNT}/class/net/${bridge}/brif/${port}" ]
}

bridge_add_port()
{
	local bridge=$1
	local port=$2

	vsw_log "Adding port ${port} to ${bridge}"

	ip_reset ${port} up

	#
	# The "master" keyword, which is required by the ip command to
	# add/remove interfaces to/from a bridge, is not yet supported on
	# iproute v2.6.32 delivered with OL6. For this reason we must use
	# the deprecated brctl command here and in bridge_remove_port()
	#
	# Using the iproute2 commands to add a port:
	# ip link set dev $port master $bridge
	#
	# To remove the port:
	# ip link set dev $port nomaster
	#

	if ! log_exec brctl addif ${bridge} ${port}
	then
		fatal_error "Failed to add port ${port} to ${bridge}"
	fi

	#ovs-vsctl add-port $bridge $port
}

bridge_remove_port()
{
	local bridge=$1
	local port=$2

	vsw_log "Removing port ${port} from ${bridge}"

	if ! log_exec brctl delif ${bridge} ${port}
	then
		fatal_error "Failed to remove port ${port} from ${bridge}"
	fi

	#ovs-vsctl del-port $bridge $port
	ip_reset ${port} down
}

vsw_name()
{
	local handle=$1

	#
	# The vswitch's name is exported in the
	# devspec file under the vswitch device file.
	#
	cat /sys/devices/channel-devices/vsw-${handle}/devspec
}

#
# Convert a macaddr from a decimal number (returned by MDPROP)
# to standard form (i.e. AA:BB:CC:DD:EE:FF) which used
# by the ip command.
#
mac_convert()
{
	local mac=$1
	local str=""

	#
	# for first five octets, add any required leading zeros, use upper case
	# and add trailing ':'
	#
	for (( CNTR=48; CNTR>0; CNTR-=8 )); do
		str+=$(printf "%.2X:" $(( (${mac} >> $CNTR) & 0xff )) )
	done

	# print final octet (with no trailing ':')
	str+=$(printf "%.2X" $((${mac} & 0xff)) )

	echo -e $str
}

check_dev_present()
{
	local interface=$1
	local tries=$MAX_ATTEMPTS

	vsw_log "Checking for device ${interface}"

	for i in $(seq ${tries}); do
		if ip link show dev ${interface} &>/dev/null; then
			vsw_log "${interface} exists on the system"
			return 0
		else
			sleep 1
		fi
	done

	return 1
}

vsw_add_bridge()
{
	local vsw=$1
	local pdev ipconf mac

	vsw_log "Adding vswitch $vsw"

	if bridge_exists $vsw; then
		vsw_log "vswitch $vsw is already present in the system"
		return 0
	fi

	# configure physical device if there's one.
	pdev=$($MDPROP -v virtual-device \
		name=virtual-network-switch,devalias=${vsw} \
		vsw-phys-dev)
	if [[ -n "${pdev}" ]]; then
		#
		# Ensure that the pdev exists on the system before attempting
		# to retrieve its configuration. Must return an error and exit
		# the script if the device is not present.
		#
		if check_dev_present ${pdev}; then
			# If the net-dev device is present, we can create the bridge
			bridge_create $vsw
		else
			fatal_error "Cannot find net-dev ${pdev}"
		fi

		ipconf=$(ip_get ${pdev})
		bridge_add_port ${vsw} ${pdev}
		ip_set ${vsw} ${ipconf}

	else
		#
		# If this is a routed bridge (i.e. no physical net-dev specified),
		# create the bridge and configure the MD assigned macaddr if there's
		# one.
		#
		bridge_create $vsw

		mac=$($MDPROP -v virtual-device \
			name=virtual-network-switch,devalias=${vsw} \
			local-mac-address)
		if [[ -n "${mac}" ]]; then
			# convert to macaddr to standard form
			mac=$(mac_convert ${mac})
			if ! log_exec ip link set dev ${vsw} address ${mac}
			then
				fatal_error "Could not configure MAC for routed bridge ${vsw}"
			fi
		fi
	fi
}

vsw_remove_bridge()
{
	local vsw=$1
	local ports port
	local ipconf

	vsw_log "Removing vswitch ${vsw}"

	if ! bridge_exists ${vsw}; then
		vsw_log "vswitch ${vsw} does not exist in the system"
		return 0
	fi

	# remove any ports on the bridge
	ports=$(bridge_list_ports ${vsw})
	for port in ${ports}
	do
		bridge_remove_port ${vsw} ${port}
		# if we removed the phys dev, move the IP to it
		if [[ "${port}" != vif* ]]; then
			ipconf=$(ip_get ${vsw})
			ip_reset ${vsw} up
			ip_set ${port} ${ipconf}
		fi
	done

	bridge_destroy ${vsw}
}

vsw_add_port()
{
	local vsw=$1
	local interface=$2

	vsw_log "Adding port ${interface} to vswitch ${vsw}"

	# Create the bridge if it doesn't exist yet
	vsw_add_bridge ${vsw}

	if bridge_port_exists ${vsw} ${interface}; then
		vsw_log "Port ${interface} is already present in vswitch ${vsw}"
	else
		bridge_add_port ${vsw} ${interface}
	fi
}

vsw_remove_port()
{
	local vsw=$1
	local interface=$2

	vsw_log "Removing vswitch ${vsw} port ${interface}"

	if bridge_port_exists ${vsw} ${interface}; then
		bridge_remove_port ${vsw} ${interface}
	else
		vsw_log "Port ${interface} is not present in vswitch ${vsw}"
	fi
}

runlock()
{
	(
		flock -x 200
		eval ${1+"$@"}
	) 200>/var/lock/vswcfg.lock
}

usage()
{
	echo "Usage: vswcfg.sh [-u]"
	echo "       vswcfg.sh [-d] <action> <vsw_name> <interface>"
	echo
	exit 2
}

main()
{
	local debug=0
	local udev=0
	local action devpath interface vswhdl

	while getopts "du" opt; do
		case "${opt}" in
			d)
				debug=1
				;;
			u)
				udev=1
				;;
			*)
				usage
				;;
		esac
	done

	#
	# If the script is invoked from udev, then we need to daemonize
	# otherwise we can deadlock when using bridge commands because
	# they can modify the device tree.
	#
	if [[ $udev -ne 0 ]]; then
		devpath=$DEVPATH
		if [[ $devpath != *vsw-port* ]]; then
			[[ $devpath =~ /devices/channel-devices/vsw-([0-9]*) ]]
			vswhdl=${BASH_REMATCH[1]}
			action=$ACTION
		else
			[[ $devpath =~ /devices/channel-devices/vsw-port-([0-9]*) ]]
			vswhdl=${BASH_REMATCH[1]}
			action=$ACTION-port
		fi
		vsw=$(vsw_name $vswhdl)
		nohup $0 $action $vsw $INTERFACE &
		exit 0
	fi

	# If the rsyslog daemon is not available when the script runs,
	# we must log to a file. Since we redirect STDERR to the LOG
	# file, we can use logger(1) with the -s option.

	if ! pidof $(which rsyslogd) &>/dev/null; then
		echo "[vswcfg]: rsyslogd not available. Redirecting output to $LOG" \
			> /dev/console
		exec 1>$LOG 2>&1
	fi

	vsw_log "args: $*"

	if [[ $debug -ne 0 ]]; then
		env
		set -x
	fi

	shift $((OPTIND - 1))

	action=$1
	vsw=$2
	interface=$3

	if [[ -z "$action" || -z "$vsw" ]]; then
		usage
	fi

	# port configuration requires an interface arg
	if [[ $action == *-port* && -z "$interface" ]]; then
		usage
	fi

	SYSFS_MNT=$(awk '$1=="sysfs" {print $2}' /etc/fstab)

	case $action in
		add)
			runlock vsw_add_bridge $vsw
			;;

		remove)
			runlock vsw_remove_bridge $vsw
			;;

		add-port)
			runlock vsw_add_port $vsw $interface
			;;

		remove-port)
			runlock vsw_remove_port $vsw $interface
			;;

		*)
			usage
			;;
	esac
}

main $@
