#!/usr/bin/python
#
# LDoms
#
# Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Represent a Machine Description (MD)

"""

import struct

MD_HEADER_FMT = "!HH I I I"
MD_ELEMENT_FMT = "!c B H I Q"

MD_HEADER_SIZE = struct.calcsize(MD_HEADER_FMT)
MD_ELEMENT_SIZE = struct.calcsize(MD_ELEMENT_FMT)

#
# MD Element Tags
#
MDE_NODE_BEGIN = "N"
MDE_NODE_END = "E"
MDE_NOOP = " "

MDE_PROP_ARC = "a"
MDE_PROP_VAL = "v"
MDE_PROP_STR = "s"
MDE_PROP_DAT = "d"

#
# MD Parser Callback Return Values
#
MD_PARSE_DONE = 0
MD_PARSE_NEXT_ELEMENT = 1
MD_PARSE_NEXT_NODE = 2


class MDParser:
    """
    Parse a Machine Description (MD)

    """

    def __init__(self):
        """
        Parse a Machine Description (MD)

        """
        pass

    def node_begin(self, name, index):
        """
        Callback invoked at the beginning of a node.

        """
        pass

    def node_end(self):
        """
        Callback invoked at the end of a node.

        """
        pass

    def prop(self, name, value):
        """
        Callback invoked for a value, string and data property.

        """
        pass

    def prop_arc(self, name, value):
        """
        Callback invoked for an arc property.

        """
        pass


class MD:
    """
    This class is used to represent a binary Machine Description (MD).

    """

    def __init__(self, md=None, mdfile=None):
        """
        Create a MD object from a file containing a binary MD.

        """
        if (md and mdfile) or ((not md) and (not mdfile)):
            raise Exception("md or mdfile should be specified")

        if md:
            self._md = md
        else:
            mdfd = open(mdfile, "r")
            self._md = mdfd.read()
            mdfd.close()

        header = struct.unpack_from(MD_HEADER_FMT, self._md)

        major = header[0]
        minor = header[1]
        if major != 1 and minor != 0:
            raise Exception("unrecognized MD version %d.%d" % (major, minor))

        self._node_blk_size = header[2]
        self._name_blk_size = header[3]
        self._data_blk_size = header[4]

        self._node_blk_begin = MD_HEADER_SIZE
        self._name_blk_begin = self._node_blk_begin + self._node_blk_size
        self._data_blk_begin = self._name_blk_begin + self._name_blk_size

    def _md_str(self, blk_begin, blk_size, data_offset, data_length):
        """
        Return data defined at the given offset in the specified
        data block. Data are returned as a single string.

        """
        if data_offset < 0 or data_offset > blk_size:
            raise Exception("invalid data offset")

        if data_length == 0:
            fmt = "s"
        else:
            fmt = "%ds" % data_length

        strval = struct.unpack_from(fmt, self._md, blk_begin + data_offset)[0]

        return strval

    def md_name(self, offset, length=0):
        """
        Return the MD name defined at the specified offset in the MD
        name table.

        """
        return self._md_str(self._name_blk_begin, self._name_blk_size,
                            offset, length)

    def md_prop_str(self, offset, length=0):
        """
        Return the string name defined at the specified offset in the MD
        data table.

        """
        strval = self._md_str(self._data_blk_begin, self._data_blk_size,
                              offset, length)
        #
        # Strip trailing NUL characters which might be included in
        # the string length.
        #
        return strval.rstrip("\0")

    def md_prop_dat(self, offset, length):
        """
        Return data name defined at the specified offset in the MD
        data table.

        """
        data = self._md_str(self._data_blk_begin, self._data_blk_size,
                            offset, length)
        return data

    def md_element(self, index):
        """
        Return the MD element at the specified index.

        """
        offset = index * MD_ELEMENT_SIZE
        if offset > self._node_blk_size:
            return None

        element = struct.unpack_from(MD_ELEMENT_FMT, self._md,
                                     self._node_blk_begin + offset)

        (tag, name_len, _unused, name_off, value) = element

        name = self.md_name(name_off, name_len)

        if tag == MDE_PROP_STR:
            voff = value & 0xffffffff
            vlen = value >> 32
            value = self.md_prop_str(voff, vlen)
        elif tag == MDE_PROP_DAT:
            voff = value & 0xffffffff
            vlen = value >> 32
            value = self.md_prop_dat(voff, vlen)

        return [tag, name, value]

    def md_parse_element(self, parser, step, index):
        """
        Parse a MD element.

        """
        if (step != MD_PARSE_NEXT_NODE and
            step != MD_PARSE_NEXT_ELEMENT):
            raise Exception("unknown step %d" % step)

        element = self.md_element(index)
        if element is None:
            return (MD_PARSE_DONE, None)

        (tag, name, value) = element

        if tag == "\0":
            return (MD_PARSE_DONE, None)

        if tag == MDE_NODE_BEGIN:
            next_step = parser.node_begin(name, index)
            if next_step == MD_PARSE_NEXT_NODE:
                # directly jump to next node
                return (next_step, value)

        elif step == MD_PARSE_NEXT_NODE:
            #
            # If we are not at the begining of a node and we want
            # the next node then skip element.
            #
            next_step = step

        elif tag == MDE_NODE_END:
            next_step = parser.node_end()

        elif (tag == MDE_PROP_VAL or
              tag == MDE_PROP_STR or
              tag == MDE_PROP_DAT):
            next_step = parser.prop(name, value)

        elif tag == MDE_PROP_ARC:
            next_step = parser.prop_arc(name, value)

        elif tag == MDE_NOOP:
            next_step = step

        else:
            raise Exception("Unexpected element %s at index %d" %
                            (tag, index))

        return (next_step, index + 1)

    def md_parse(self, parser):
        """
        Parse a MD. Invoke parser callbacks for each element of the MD.

        """
        index = 0
        step = MD_PARSE_NEXT_ELEMENT

        while step != MD_PARSE_DONE:
            (step, index) = self.md_parse_element(parser, step, index)
            if step is None:
                # None means continue
                step = MD_PARSE_NEXT_ELEMENT
