#!/usr/bin/python
#
# LDoms
#
# Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
# The Universal Permissive License (UPL), Version 1.0
#
# Subject to the condition set forth below, permission is hereby
# granted to any person obtaining a copy of this software, associated
# documentation and/or data (collectively the "Software"), free of
# charge and under any and all copyright rights in the Software, and
# any and all patent rights owned or freely licensable by each licensor
# hereunder covering either (i) the unmodified Software as contributed
# to or provided by such licensor, or (ii) the Larger Works (as defined
# below), to deal in both
#
# (a) the Software, and
# (b) any piece of software and/or hardware listed in the lrgrwrks.txt
# file if one is included with the Software (each a "Larger Work" to
# which the Software is contributed by such licensors),
#
# without restriction, including without limitation the rights to copy,
# create derivative works of, display, perform, and distribute the
# Software and make, use, sell, offer for sale, import, export, have
# made, and have sold the Software and the Larger Work(s), and to
# sublicense the foregoing rights on either these or other terms.
#
# This license is subject to the following condition: The above
# copyright notice and either this complete permission notice or at a
# minimum a reference to the UPL must be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Framework for decoding a MD node with MDDecoder.

"""

from ldoms.MD import MD
from ldoms.MDDecoder import MDDecoder


class MDNodeDecoder:

    """
    Framework for decoding a MD node with MDDecoder.

    """

    def __init__(self, node_name, node_index):
        pass

    def begin(self):
        """
        Callback invoked at the beginning of a node.

        """
        pass

    def end(self):
        """
        Callback invoked at the end of a node.

        """
        pass

    def prop(self, prop_name, prop_value):
        """
        Callback invoked for a value, string and data property.

        """
        pass

    def arc(self, arc_name, node_index):
        """
        Callback invoked for an arc property.

        """
        pass

    def finalize(self, decoder):
        """
        Callback invoke after the decoding has been completed.

        """
        pass


class MDNodeDictDecoder(MDNodeDecoder):

    """
    Implement the MDNodeDecoder class to decode MD nodes into dict.

    """

    def __init__(self, node_name, node_index):
        MDNodeDecoder.__init__(self, node_name, node_index)
        self.node_name = node_name
        self.node_index = node_index
        self.object = {}
        self.object_arcs = []

    def end(self):
        return self.object

    def prop(self, prop_name, prop_value):
        self.object[prop_name] = prop_value

    def arc(self, arc_name, node_index):
        if arc_name == 'fwd':
            self.object_arcs.append(node_index)

    def finalize(self, decoder):
        for node_index in self.object_arcs:
            peer_node = decoder.node_resolve(node_index)
            if peer_node is None:
                continue
            peer_name = peer_node.node_name
            self.object.setdefault(peer_name, [])
            self.object[peer_name].append(peer_node.object)


def md_node_dict_decoder_factory(factory_rules, node_name, node_index):
    """
    Create a MDNodeDictDecoder for a node if the node name matches the
    factory rules.

    """
    if node_name in factory_rules:
        node_decoder = MDNodeDictDecoder(node_name, node_index)
    else:
        node_decoder = None

    return node_decoder


def md_node_list_collector(result, node_name, node_object):
    """
    Collect nodes into lists based on the node name.

    """
    node_list = result.setdefault(node_name, [])
    node_list.append(node_object)


def md_decode_nodes(node_names, md=None, mdfile=None):
    """
    Decode MD nodes with the specified node names.

    """
    if (not md) and (not mdfile):
        mdfile = "/dev/mdesc"
    mdesc = MD(md=md, mdfile=mdfile)
    nodes = {}
    decoder = MDDecoder()
    decoder.decode(mdesc,
                   md_node_dict_decoder_factory, node_names,
                   md_node_list_collector, nodes)
    return nodes
